// ======================================================================
//
// direct3d_5.cpp
// jeff grills
//
// copyright 1997 origin systems inc
//
// ======================================================================

#include "first.h"
#include "Direct3d_5.h"

#include "DirectDraw.h"
#include "MaterialData_d3d_5.h"
#include "ConfigPlatform.h"

#include <float.h>
#include <stdio.h>
#include <string.h>

// ======================================================================

LPDIRECT3D2       Direct3d::d3d;
Direct3d::Driver *Direct3d::firstDriver;
Direct3d::Driver *Direct3d::lastDriver;
int               Direct3d::globalDriverId;

// ======================================================================

void Direct3d::enumerate(void)
{
	int                       i;
	HRESULT                   result;
	LPDIRECTDRAW              ddraw;
	LPDIRECT3D2               d3d;
	const DirectDraw::Driver *ddd;

	if (!firstDriver)
	{
		DirectDraw::enumerate();

		for (i = 1; (ddd = DirectDraw::getDriver(i)) != NULL; ++i)
		{
			result = DirectDrawCreate(ddd->pGuid, &ddraw, NULL);
			FATAL_HR("d3d::enumerate create dd %d", result);

			result = ddraw->QueryInterface(IID_IDirect3D2, reinterpret_cast<void **>(&d3d));
			FATAL_HR("d3d::enumerate qi d3d %d", result);

			// get a list of rendering devices
			result = d3d->EnumDevices(enumDevicesCallback, const_cast<void *>(reinterpret_cast<const void *>(ddd)));
			FATAL_HR("d3d::enumerate enumDevices %d", result);
			
			d3d->Release();
			ddraw->Release();
		}
	}
}

// ----------------------------------------------------------------------

HRESULT WINAPI Direct3d::enumDevicesCallback(LPGUID lpGuid, LPSTR lpDeviceDescription, LPSTR lpDeviceName, LPD3DDEVICEDESC lpHWDesc, LPD3DDEVICEDESC lpHELDesc, LPVOID lpContext)
{
	Driver      *driver;

	// create a new driver entry
	driver = NEW Driver;

	driver->id = ++globalDriverId;
	driver->directDrawDriver = reinterpret_cast<const DirectDraw::Driver *>(lpContext);

	// add driver to linked list
	driver->next = NULL;
	if (!firstDriver)
		firstDriver = driver;
	else
		lastDriver->next = driver;
	lastDriver = driver;

	// copy data
	driver->guid = *lpGuid;
	driver->name = strdup(lpDeviceName);
	driver->desc = strdup(lpDeviceDescription);
	driver->hw   = *lpHWDesc;
	driver->sw   = *lpHELDesc;

	// log the device
	MESSAGE(true, ("d3d::enumDev %d=d3d %d=dd '%s' '%s'\n", driver->id, driver->directDrawDriver->id, driver->name, driver->desc));

	// figure some useful things out about the device
	driver->isHardware        = (lpHWDesc->dcmColorModel) != 0;
	driver->primary           = (driver->isHardware ? &driver->hw : &driver->sw);
	driver->canTexture        = (driver->primary->dpcTriCaps.dwTextureCaps & D3DPTEXTURECAPS_PERSPECTIVE) != 0;
	driver->zBuffer           = (driver->primary->dwDeviceZBufferBitDepth != 0);
	driver->rgbLighting       = (driver->primary->dcmColorModel == D3DCOLOR_RGB);
	driver->dithering         = (driver->primary->dpcTriCaps.dwRasterCaps & D3DPRASTERCAPS_DITHER);
	driver->alphaBlending     = (driver->primary->dpcTriCaps.dwSrcBlendCaps & D3DPBLENDCAPS_SRCALPHA) != 0 && (driver->primary->dpcTriCaps.dwDestBlendCaps & D3DPBLENDCAPS_INVSRCALPHA) != 0;
	driver->bilinearFiltering = (driver->primary->dpcTriCaps.dwTextureFilterCaps & D3DPTFILTERCAPS_LINEAR) != 0;
	driver->stippledAlpha     = (driver->primary->dpcTriCaps.dwShadeCaps & D3DPSHADECAPS_ALPHAFLATSTIPPLED) != 0 && (driver->primary->dpcTriCaps.dwShadeCaps & D3DPSHADECAPS_ALPHAFLATBLEND) == 0;

	return D3DENUMRET_OK;
}

// ----------------------------------------------------------------------

const Direct3d::Driver *Direct3d::getDriver(int id)
{		
	const Driver *driver;

	// find the specified driver
	for (driver = firstDriver; driver && driver->id != id; driver = driver->next)
		;

	return driver;
}

// ======================================================================

void Direct3d::install(const Driver *driver)
{
	HRESULT              result;
	LPDIRECTDRAW2        ddraw;

#if USE_FPU_24 == 1
	// reset FPU to 53 bits of precision
	int                  origFPU;
	origFPU = _control87(0,0);
	_control87(_PC_53, _MCW_PC);
#endif

	MESSAGE(true, ("Direct3d::install using %d '%s' '%s'\n", driver->id, driver->name, driver->desc));
	MESSAGE(true, ("Direct3d::install   %d=hardware %d=texture %d=zBuffer %d=rgbLight\n", static_cast<int>(driver->isHardware), static_cast<int>(driver->canTexture), static_cast<int>(driver->zBuffer), static_cast<int>(driver->rgbLighting)));
	MESSAGE(true, ("Direct3d::install   %d=dither %d=alphaBlend %d=bilinear %d=stippleAlpha\n", static_cast<int>(driver->dithering), static_cast<int>(driver->alphaBlending), static_cast<int>(driver->bilinearFiltering), static_cast<int>(driver->stippledAlpha)));

	DEBUG_FATAL(d3d, ("Direct3d::install already"));

	ddraw = DirectDraw::getDirectDrawObject();
	DEBUG_FATAL(!ddraw, ("Direct3d::install !ddraw"));

	// create direct3d object
	result = ddraw->QueryInterface(IID_IDirect3D2, (LPVOID*)&d3d);
	FATAL_HR("d3d::install create %d", result);

#if USE_FPU_24 == 1
	_control87(origFPU, _MCW_PC);
#endif
}

// ----------------------------------------------------------------------

void Direct3d::remove(void)
{
	HRESULT    result;
	Driver *driver;

	// whack the enumerated device drivers
	while (firstDriver)
	{
		driver = firstDriver;
		firstDriver = firstDriver->next;

		delete [] driver->name;
		delete [] driver->desc;
		delete driver;
	}
	firstDriver = lastDriver = NULL;
		
	if (d3d)
	{
		// -q- there should still be a refcount of 1 after this because the DD object for D3D still exists
		result = d3d->Release();
		d3d = NULL;
	}
}

// ----------------------------------------------------------------------


// ======================================================================
